/* proxy.c - guts of the telnet proxy */
/* Created: Sat Mar  6 08:06:50 AM EST 2021 malakai */
/* $Id: proxy.c,v 1.4 2021/03/12 22:12:26 malakai Exp $ */

/* Copyright © 2021 Jeff Jahr <malakai@jeffrika.com>
 *
 * This file is part of MUDitM - MUD in the Middle
 *
 * MUDitM is free software: you can redistribute it and/or modify it under the
 * terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * MUDitM is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with MUDitM.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <arpa/inet.h>
#include <arpa/telnet.h>
#include <fcntl.h>
#include <malloc.h>
#include <netdb.h>
#include <netinet/in.h>
#include <openssl/err.h>
#include <openssl/ssl.h>
#include <poll.h>
#include <signal.h>
#include <stdio.h>
#define __USE_GNU
#include <string.h>
#include <unistd.h>

#include "debug.h"
#include "muditm.h"
#include "iobuf.h"

#include "proxy.h"

struct endpoint_data *new_endpoint() {
	struct endpoint_data *ep;

	ep = (struct endpoint_data *)malloc(sizeof(struct endpoint_data));
	ep->socket = -1;
	ep->ssl = NULL;
	ep->direction = EP_UNSPEC;
	ep->mnes_state = 0;
	ep->iobuf[EP_INPUT] = new_iobuf(EP_BUFSIZE);
	ep->iobuf[EP_OUTPUT] = new_iobuf(EP_BUFSIZE);
	return(ep);

}

int close_endpoint(struct endpoint_data *ep) {

	int ret;

	if(ep->ssl != NULL) {
		SSL_shutdown(ep->ssl);
		SSL_free(ep->ssl);
		ep->ssl = NULL;
	}
	if(ep->socket >= 0) {
		ret = close(ep->socket);
		ep->socket = -1;
	}
	return(ret);
}

void free_endpoint(struct endpoint_data *ep) {
	if(!ep) return;
	close_endpoint(ep);
	if(ep->iobuf[EP_INPUT]) free_iobuf(ep->iobuf[EP_INPUT]);
	if(ep->iobuf[EP_OUTPUT]) free_iobuf(ep->iobuf[EP_OUTPUT]);
	free(ep);
}

char *addr_endpoint_bleh(struct endpoint_data *ep, char *buf, size_t size) {

	struct sockaddr_storage sa;
	socklen_t sa_size;

	sa_size = sizeof(sa);
	*buf='\0';

	if(getsockname(ep->socket,(struct sockaddr *)&sa, &sa_size) <0) {
		muditm_log("getsockname: ",strerror(errno));
		return(buf);
	}
	
	switch(sa.ss_family) {
		case AF_INET6:
			inet_ntop(AF_INET6,&((struct sockaddr_in6 *)&sa)->sin6_addr,buf,size);
			break;
		case AF_INET:
			inet_ntop(AF_INET,&((struct sockaddr_in *)&sa)->sin_addr,buf,size);
			break;
		default:
			break;
	}
	return(buf);
}

char *addr_endpoint(struct endpoint_data *ep, char *buf, size_t size) {

	struct sockaddr_in6 sa;
	socklen_t sa_size = sizeof(struct sockaddr_in6);

	*buf='\0';

	if(getsockname(ep->socket,(struct sockaddr *)&sa, &sa_size) <0) {
		muditm_log("getsockname: ",strerror(errno));
		return(buf);
	}
	
	inet_ntop(AF_INET6,&((struct sockaddr_in6 *)&sa)->sin6_addr,buf,size);
	return(buf);
}

ssize_t read_endpoint(struct endpoint_data *ep, void *buf, size_t num) {

	int ret, err;

	if(ep->ssl) {
		while(1) {
			ret = SSL_read(ep->ssl,buf,num);
			if(ret == -1) {
				err = SSL_get_error(ep->ssl,ret);
				if(err == SSL_ERROR_WANT_READ) {
					continue;
				}
			}
			break;
		}
	} else {
		return(read(ep->socket,buf,num));
	}
	return(ret);
}

ssize_t write_endpoint(struct endpoint_data *ep, void *buf, size_t num) {

	int ret, err;

	if(ep->ssl) {
		while(1) {
			ret = SSL_write(ep->ssl,buf,num);
			if(ret == -1) {
				err = SSL_get_error(ep->ssl,ret);
				if(err == SSL_ERROR_WANT_WRITE) {
					continue;
				}
			}
			break;
		}
	} else {
		/* fixme... ewouldblock, eagain... */
		ret = write(ep->socket,buf,num);
	}
	return(ret);
}

int ssl_start_endpoint(struct endpoint_data *ep, SSL_CTX *ctx) {

	int ret;

	if (!ctx) {
		muditm_log("Missing SSL context!");
		ERR_print_errors_fp(stderr);
		exit(EXIT_FAILURE);
	}

	ep->ssl = SSL_new(ctx);
	SSL_set_fd(ep->ssl,ep->socket);

	switch (ep->direction) {
		case EP_CLIENT:
			muditm_log("SSL start client mode on socket %d",ep->socket);
			if ( (ret=SSL_accept(ep->ssl)) <= 0) {
				ERR_print_errors_fp(stderr);
				return(ret);
			} 
			break;
		case EP_SERVER:
			muditm_log("SSL start server mode on socket %d",ep->socket);
			if ( (ret=SSL_connect(ep->ssl)) <= 0) {
				ERR_print_errors_fp(stderr);
				return(ret);
			} 
			break;
		case EP_UNSPEC:
		default:
			muditm_log("SSL allocated but not started for socket %d",ep->socket);
			ret = 1;
	}
	return(ret);
}


/* memmem_p() - like memmem() but will return pointer to a partial match, and
 * sets *part to the length of the partial match.  If called with non-null,
 * non-zero part, search contiues from that location of needle, and must be
 * found at the start of the haystack, or call returns NULL and leaves part
 * untouched. */

char *memmem_p(char *haystack, size_t haystacklen, char *needle, size_t needlelen, size_t *part) {

	char *h, *hend;
	char *n, *nend;
	char *m;
	size_t localpart = 0;

	if(haystacklen == 0 || needlelen ==0) {
		return(NULL);
	}

	if(!part) {
		part = &localpart;
	}

	hend = haystack+haystacklen;
	nend = needle+needlelen;
	m = NULL;
	n = needle + (*part);

	for(h=haystack; h<hend && n<nend; h++) {
		if(*h == *n) {
			if(!m) m = h;
			n++;
		} else {
			m = NULL;
			if(*part) {
				break;
			} else {
				n = needle + (*part);
			}
		}
	}
	if(m) {
		*part = n-needle;
	}
	return(m);

}

int snprintf_mnes_pair(char *str, size_t size, char *var, char *val) {
	int s;
	s=snprintf(str,size,"%c%c%c%c%c%s%c%s%c%c",
		IAC, SB, TELOPT_NEW_ENVIRON, TELQUAL_IS,
		NEW_ENV_VAR, var,
		NEW_ENV_VALUE, val,
		IAC, SE
	);
	return(s);
}

int muditm_proxy(struct endpoint_data *client, struct endpoint_data *game) {

	struct pollfd pollster[2];
	struct flow_data flow[2];
	int pollster_count = 2;
	int polltimeout = 1000;
	int ready;
	ssize_t bytes_recv, bytes_sent;
	char *match;
	size_t match_len;
	char addrtxt[INET6_ADDRSTRLEN];

	struct iobuf_data *iob,*out;

	char mnes_trigger[] = { IAC, SB, TELOPT_NEW_ENVIRON, TELQUAL_SEND, IAC, SE };
	char mnes_wont[] = { IAC, WONT, TELOPT_NEW_ENVIRON };

	pollster[0].fd = client->socket;
	pollster[0].events = POLLIN;
	flow[0].in = client;
	flow[0].out = game;
	
	if(fcntl(client->socket,F_SETFL, O_NONBLOCK) == -1) {
		muditm_log("Couldn't set client side to non-blocking io mode: ",strerror(errno));
		return(-1);
	}

	pollster[1].fd = game->socket;
	pollster[1].events = POLLIN;
	flow[1].in = game;
	flow[1].out = client;

	if(fcntl(game->socket,F_SETFL, O_NONBLOCK) == -1) {
		muditm_log("Couldn't set server side to non-blocking io mode: ",strerror(errno));
		return(-1);
	}

	while(1) {

		/* poll for input */
		ready = poll(pollster,pollster_count,polltimeout);

		if(ready == -1) {
			muditm_log("Polling error: %s",strerror(errno));
			return(-1);
		}

		if(ready == 0) {
			//muditm_log("Idle Tick...");
			continue;
		}

		for(int i=0;i<pollster_count;i++) {
			if(pollster[i].revents & POLLIN) {
				iob = (flow[i].in->iobuf[EP_INPUT]);
				bytes_recv = read_endpoint(flow[i].in,tail_iobuf(iob),avail_iobuf(iob));
				if(bytes_recv == -1) {
					/* there was an error. */
					return(-1);
				}	
				if(bytes_recv == 0) {
					/* He hung up. */
					muditm_log("%s has closed the connection.",
						(flow[i].in->direction==EP_CLIENT)?"Client":"Game"
					);
					return(1);
				}	
				push_iobuf(iob,bytes_recv);

				/* This is the creamy filling in the middle */
				
				if(flow[i].in->direction == EP_SERVER) {

					while(1) {
						match_len = 0;
						match = memmem_p(head_iobuf(iob), len_iobuf(iob),
							mnes_trigger, sizeof(mnes_trigger), 
							&(match_len)
						);

						if(match) {
							if(match_len == sizeof(mnes_trigger)) {
								muditm_log("Server mnes trigger full match!");
								if(flow[i].out->mnes_state == WONT) {
									/* send all the bytes up to the match, but
									 * dont forward the match itself */
									bytes_sent = write_endpoint( flow[i].out,
										head_iobuf(iob),
										(match - head_iobuf(iob))
									);
									pop_iobuf(iob,match-head_iobuf(iob)+match_len);
								} else {
									/* send all the bytes up to and including the match */
									bytes_sent = write_endpoint(flow[i].out,
										head_iobuf(iob), 
										(match+match_len - head_iobuf(iob))
									);
									pop_iobuf(iob,(match+match_len)-head_iobuf(iob));
								}

								/* send some variables. */
								out = (flow[i].in->iobuf[EP_OUTPUT]);
								push_iobuf(out, 
									snprintf_mnes_pair(tail_iobuf(out),avail_iobuf(out),
										"PROXY_NAME",muditm_proxy_name
									)
								);
								addr_endpoint(client,addrtxt,sizeof(addrtxt));
								push_iobuf(out, 
									snprintf_mnes_pair(tail_iobuf(out),avail_iobuf(out),
										"IPADDRESS",addrtxt
									)
								);
								write_endpoint(flow[i].in,
									head_iobuf(out),
									len_iobuf(out)
								);

								/* Great!  go back and try matching again against what's left */
								continue;
							} else {
								muditm_log("Server mnes trigger partial match %d bytes...",match_len);
								/* done, still needing to add more input. */
								break;
							}
						} else {
							/* write the whole buffer */
							bytes_sent = write_endpoint(flow[i].out,head_iobuf(iob),len_iobuf(iob));
							popall_iobuf(iob);

							/* done, all available input is processed! */
							break;
						}
					}

				} else if(flow[i].in->direction == EP_CLIENT) {
					//bytes_sent = write_endpoint(flow[i].out,head_iobuf(iob),len_iobuf(iob));
					//pop_iobuf(iob,len_iobuf(iob));

					while(1) {
						match_len = 0;
						match = memmem_p(head_iobuf(iob), len_iobuf(iob),
							mnes_wont, sizeof(mnes_wont), 
							&(match_len)
						);

						if(match) {
							if(match_len == sizeof(mnes_wont)) {
								muditm_log("Client mnes wont trigger full match!");
								/* note that the client wont be doing mnes. */
								flow[i].out->mnes_state = WONT;
								/* ...but we're going to lie to the server about it. */
								*(match+1)= WILL;
								/* send all the bytes up to and including the match */
								bytes_sent = write_endpoint(flow[i].out,
									head_iobuf(iob), 
									(match+match_len - head_iobuf(iob))
								);
								pop_iobuf(iob,(match+match_len)-head_iobuf(iob));
								/* Great!  go back and try matching again against what's left */
								continue;
							} else {
								muditm_log("Server mnes wont partial match %d bytes...",match_len);
								/* done, still needing to add more input. */
								break;
							}
						} else {
							/* write the whole buffer */
							bytes_sent = write_endpoint(flow[i].out,head_iobuf(iob),len_iobuf(iob));
							popall_iobuf(iob);

							/* done, all available input is processed! */
							break;
						}
					}

				}
				
			}
		}
	}
}

